#sv.py
# Functions to support DVR+ Services and ServiceDetails calls
#
# Copyright June 2020, Timothy Goihl
#
# Handle the 2 'service' request messages from the DVR+

from datetime import datetime
import json
import copy
import logging
from sd import SDGetStationMapping
from urllib.parse import urlparse  #, parse_qs
import requests

import os
from ut import LocalTimeStr
from ut import PigsPath


# DVR+ Message type 1 - Return the service code when given a ZIP code
# e.g. 20499 is the service ID for zip=53593
# Most of the structure does not appear to be used.
# The key return data is the ServiceId -
# it must match throughout responses, in service details, and in channel listings
# We are going to use 99999 for anything we send to DVR+

# This structure was captured from CM server to DVR+
# this is a Python Dictionary

SR = {
  "ServicesResult": {
    "RequestId": "4db8c3b3-1096-43e8-8fa8-f64918f5ddb7",
    "TimeStamp": "2020-05-02T14:25:24.104Z",
    "Status": "PNE",  # means Processed No Errors, NPE Not Processed, Errors
    "Errors": [],
    "EndTimestamp": "0001-01-01T00:00:00",
    "Build": "Development-v9-dev-9.15.08-173-r34606-34606"
    #"Services": ServicesDef
  }
}

ServicesDef =  {
      "Service": {
        "ServiceClass": "Over the Air Broadcast",
        "ServiceId": "20499",     # We would change this to 99999 and send it back
        "Name": "Madison Area Channels - Broadcast TV",
        "City": "Madison",
        "Type": "Broadcast",
        "MSO": "Broadcast TV",
        "MSOID": "306430",
        "SystemName": "Broadcast TV",
        "TimeZones": [
          {
            "StartDateTime": "2020-03-08T08:00:00Z",
            "EndDateTime": "2020-11-01T07:00:00Z",
            "Offset": "-300"
          }
        ]
      }
    }  # end services

# return a response that will make DVR+ happy.
# We are using the json structure captured with WireShark
# We just update the time, insert our ServiceID of 99999
# and send it to DVR+

def GetServiceIdFromZip(theZip, serviceID, tzlist):
                         
    theTime=datetime.utcnow()

    #%f is microsecs.  need mS?  setting to '000' for now
    SR["ServicesResult"]["TimeStamp"] = theTime.strftime("%Y-%m-%dT%H:%M:%S.000Z")
    
    SR["ServicesResult"]["Services"]  = ServicesDef

    SR["ServicesResult"]["Services"]["Service"]["ServiceId"]= serviceID
    SR["ServicesResult"]["Services"]["Service"]["TimeZones"]= copy.deepcopy(tzlist)

    # ToDo Testing to generate DVR+ error codes
    #ErrStr = {
    #    "ErrorCode": "6000",
    #    "ErrorName": "Service_SystemException",
    #    "ErrorText": "System exception, Please contact system administrator (Server: 09)",
    #    "ErrorType": "Error"
    #    }
    #SR["ServicesResult"]["Errors"].clear()
    #SR["ServicesResult"]["Errors"].append(ErrStr)
    #SR["ServicesResult"]["Status"] = "NPE"

    #convert Python dict to json
    js = json.dumps(SR,indent=2)
    
    return js

# Message type 2 - Return the service Details when given a a service id code
# The DVR+ takes the ServiceID obtained from the zip code
# and requests the 'channel lineup' for that ServiceID
# The key data returned by the CM server is a list of Channels[].  SD calls them StationIDs
# We know DVR+ uses the channel images from this response

# This is a Python Dictionary
# These are the captured keys from a CM to DVR+ guide refresh
# commenting out as many of the fields as possible to minimize data transfer and size
# if I comment out too many, it doesnt work
ServiceDetailHeader = { 
  "ServiceDetailsResult": {
    "RequestId": "6c7838be-e7ab-4a26-8ef4-1a9da20f27e1",
    "TimeStamp": "2020-05-02T15:46:17.839Z",  #update this when sending
    "Status": "PNE",
    "Errors": [],
    "EndTimestamp": "0001-01-01T00:00:00",
    "Build": "Development-v9-dev-9.15.08-173-r34606-34606",
    "ChannelLineup": {
      "ServiceId": 20499,    #update this when sending
      "ServiceClass": "Over the Air Broadcast",
      "Name": "Madison Area Channels - Broadcast TV",
      "City": "Madison",
      "Type": "Broadcast",
      "MSO" : "Broadcast TV",
      "MSOID": "306430",
      "TimeZones": [
        {
          "StartDateTime": "2020-03-08T08:00:00Z",
          "EndDateTime": "2020-11-01T07:00:00Z",
          "Offset": -300
        }
      ],
      "Channels": []    # ChannelDef  #create this list when sending
    }
  }
}

# Definition of a Channel - SD calls a channel a station
# The Channel must match the VirtualChannelNumber or DVR+ will not display listings
# The SourceId here must match the SourceId in the Airing or DVR+ will not display listings

# SchedulesDirect to DVR+ mapping
#  "SourceId": 11833,              SD StationID from Lineup/Stations must match airings SourceId
#                                  must match channel to atscMajor.atscMinor from StationMap
#  "Channel": "3.1",               SD atscMajor.atscMinor from Lineup/map
#  "VirtualChannelNumber": "3.1",  SD atscMajor.atscMinor from Lineup/map
#  "CallLetters": "WISCHD",        SD affiliate from Lineup/Stations
#  "DisplayName": "WISCHD",        SD callsign from Lineup/Stations
#  "FullName": "WISC HDTV",        SD name from Lineup/Stations
#  "IconAvailable": True,          SD logo URL from Lineup/Stations

# these are the captured keys from a CM to DVR+ guide refresh
# commenting out as many of the fields as possible to minimize data transfer and size
# if I comment out too many, it doesnt work
# This is a Python Dictionary

ChannelDef = {          
  #fields we have learned are important for the DVR+ schedule
    
  "SourceId": 11833,        # must match airings SourceId StationID
  "Channel": "3.1",         # must match in each program listing
  "CallLetters": "WISCHD",  # Must match CallLetters is Prog Airing 18chars
  "DisplayName": "WISCHD",  # Must match SourceDisplayName in Prog Airing 30c
  "FullName": "WISC HDTV",  # Match to SourecLong Name in Prog Airing
  "VirtualChannelNumber": "3.1",  # must match in 'Channel'
  
  "SDLineupName" : "",      # Since this is json, I should be able add my own Keys
 
  # I dont have a way to look these up.  What to put here???
  
  "Affiliation": 0,         # is this important?  was 2165

  "ParentNetworkId": 2,     # must match program airing listings. same network same logo
  "IconAvailable": True,    # true if icon available for PartentNetworkId

  # do these do anything???
  "Order": 1,              # preferred sort order.  I will orderthem 1,2,3,...
  "SourceType": "Broadcast",

  # other fields, do we need these?  it didnt work with them commented too many out
  "ChannelLocation": "",   #c-band location????   *** PiGS will use this for the uhfVhf RF channel  ***
  "DvbTriplets": [],       #eurpean
  "Type": "24-Hours",      #"Partial","Shared"
  "ChannelSchedules": [],  # broadcast times if not 24-hours
  
  "TiVoSupported": True,
  "ServiceType": "Digital",
  #  "SourceAttributes": "1",  # number of attributes for find in next list
  "SourceAttributes": "1",  # bit mask 1=HD,2=Music,4=PPV,8=Stet (govaccess)
  "SourceAttributeTypes": [
    "HD"
  ],
  "ChannelImages": []    #ImageDef, link to info below
}

# these are the captured keys from a CM to DVR+ guide refresh
# commenting out as many of the fields as possible to minimize data transfer and size

ImageDef =  {
#  "ImageUrl": "http://cps-static.rovicorp.com/2/Open/92x36_352/Source/253/wisctv.png?partner=mikecrabtreev4edzwwtca4e847n72j8kesw",
  "ImageUrl": "http://epg.channelmastertv.com/pigoutline92x36.png",  # this is a PiGS icon
  "ObjectId": "11833",  # must match SourceId above to display
  "ObjectType": "Source",
  "ObjectName": "WISC HDTV",
  
  "ImageUseType": None,
  
  "ImageId": 427273405,          # not changing this, dvr+ doesnt seem to care
  "ImageType": "Station Logo",
  "ImageTitle": "Pig Outline",
  
  "ImageCaption": "92x36",
  "Relevancy": True,
  "ImageCreditDisplay": None,
  "ImageCastWeight": None,
  "ImageZoomLevel": "1",
  "ImageHeadcount": None,
  "ImageFormat": None,
  "ImageHorizontalResolution": 92,
  "ImageVerticalResolution": 36,
  "ImageMaintainAspectFlag": None,
  "ImageExpiryDateTime": None,
  "LastUpdate": "2019-02-06T22:36:38.001Z",
  "ImageFormatId": 0,
  "Orientation": "Landscape",
  "AspectRatio": "23:9",
  "ParentImageId": 28473370,
  "MaxImageScale": "100"
}

# build the channel services list at boot and occasionally when needed (UI Save Settings)
# this will combine info from vairous sources into one common list of channel info
# This list is formatted for DVR+ so dont mess with it too much

# for pigstatus
ServicesGrpHeader = 'Services List'
CHAN_KEY = "Channels"
BUILDKEY = "BuildTime"

SvcBuildInfo = {}
SvcBuildInfo[ServicesGrpHeader] = {}

def GetSVCBuildInfo() :
    return SvcBuildInfo
    
gServiceList = []   # a list of channelDef objects, the 'service' for area

def GetServiceList() :
    global gServiceList
    return gServiceList


from ut import utSaveObjectToDisk, utLoadObjectFromDisk

# SaveLevel: 0=Dont Save, 1=SaveCache, 2=Save Boths Progs & Cache

SERVICE_FILE = 'ServicesList'

def SVLoadServicesFromDisk() :
    
    global gServiceList
    global SvcBuildInfo
    
    svcList = utLoadObjectFromDisk(SERVICE_FILE)
    
    if svcList :
        gServiceList = svcList
        
        # Set last built time based on file modified(saved) time
        psf = PigsPath(SERVICE_FILE+'.pkl')
        uts = os.path.getmtime(psf)  # file times are utc unix timestamps
        lbt = datetime.utcfromtimestamp(uts)
        lt,lz = LocalTimeStr(lbt, showSeconds=True)

        # update module status
        SvcBuildInfo[ServicesGrpHeader][BUILDKEY] = lt
        SvcBuildInfo[ServicesGrpHeader][CHAN_KEY] = len(gServiceList)


#Python Version        3.9 3.8 3.7 3.6 3.5 3.4
#Pillow >= 8.0         Yes Yes Yes Yes     
#Pillow 7.0 - 7.2          Yes Yes Yes Yes 
#Pillow 6.2.1 - 6.2.2      Yes Yes Yes Yes 
#Pillow 6.0 - 6.2.0            Yes Yes Yes 
#Pillow 5.2 - 5.4              Yes Yes Yes Yes

# This will build a the list of channels, called the service list
# list is built using the SD Lineups in MyLineups[] from json config file -
# It needs the LINEUP channel map for station ID's
# It will be kept in memory in gServiceList=[] global list
# It only needs to update if the user changes their lineup, favorites chans,
#  or maybe if new channels are added, or a repack type thing is done.
# The gServiceList dict is built each time PiGS boots, or when UI clicks SaveSettings

from cm import CMGetServicesFromPostalCode
from cm import CMDisposeServcies
from cm import CMGetChannelImageDef
from ut import VersionStringToFloat


ICON_NONE = 0
ICON_SCHD = 1
ICON_ROVI = 2

def BuildServicesListNEW(Settings, ServiceID, token):
    
    global gServiceList
    global SvcBuildInfo

    from io  import BytesIO

    logging.info("BSL: Building Services List V2. Token=%s", token)
    logging.info("BSL: IconType(%s), IconTheme(%s)", Settings['IconType'], Settings['SdIconTheme'])
    if Settings["UpdateIconCache"] :
        logging.info("BSL: Icon Cache Update requested.  UpdateIconCache=%s", Settings["UpdateIconCache"])
    
    gServiceList.clear()  # python 3.0, clear the list

    PillowLib = True
    PillowVer = 0.0

    try :
        import PIL
        VerStr = PIL.__version__
        PillowVer = VersionStringToFloat(VerStr)
        logging.info("BSL: Pillow Library version %s", PillowVer)
        
        #if PillowVer
        
        from PIL import Image
                
    except:
        logging.warning("BSL: Could Not Import Pillow Image Library")
        logging.warning("     maybe try?  sudo apt-get install Pillow")
        logging.warning("     maybe try?  sudo apt-get install libopenjp2-7")
        PillowLib = False

    # Status Grouping Title key
    SvcBuildInfo[ServicesGrpHeader] = {}

    # Cycle through the users lineup & Favorite Stations list and fill in the channels
    #https://www.w3schools.com/python/python_dictionaries.asp
    
    # Get the Users Favorites list
    # http://json.schedulesdirect.org/20141201/lineups/USA-OTA-53593
    
    # check for empty list
    if not Settings['MyLineups'] : 
        logging.warning('BSL: There are no Lineups!\n')
        return

    order=1  # we will order the channels to make DVR+ happy

    ChannelList = []   # watch for duplicate channels

    # loop and add Channels / station data
    # Do this each time in case user has changed thier lineup
    # Each lineup is equivalent to a postal code, we can do unlimited lineups,
    # but SD limits a user to 4 lineups
    
    for LU in Settings['MyLineups'] :  #  loop thru the Users lineups in the Config File
        
        logging.info("BSL: Processing LU %s", LU['LineupName'])
        
        # Get the Mapping of Lineup to Channels #'s to StationId's
        httpcode, Maps = SDGetStationMapping( LU['LineupName'] )

        if httpcode != 200 :
            logging.warning("BSL: LU %s, Error code = %d", LU['LineupName'], httpcode)
            continue    # lineup probably not in user account

        # if user wants to use the old CM channel icons, we will look them up
        # the CM module will cache the ServiceList with icon urls and all.
        # Get the CM services for the current zipcode
                
#        if Settings['UseCMImages'] :
        if Settings['IconType'] == ICON_ROVI:
            split = LU['LineupName'].split('-')  # split off the zipcode, US or CAN
            postal = split[-1]   # the last item in the created list is the ZIP
            logging.info("BSL: Fetching CM Images links for (%s)", postal )    
            code = CMGetServicesFromPostalCode(postal)

        # base of 2 lists in SD dictionary object
        # for some dumb? reason, there are two parallel structures for each station, but each has info we need
        
        mapList = Maps["map"]      # stationID, uhfVhf, atscMajor,atscMinor
        
        # stationsList has some of stuff we could use in the future - does not have Ch num like 3.1
        # have to lookup and match stationID str
        
        stationList = Maps["stations"] # stationID, name, callsign, affiliate, images, and MORE

        # From the LineupPreview we did on the UI, we have most of the Info needed
        # Really we just need to lookup the stationID from the stationList 

        # Tested adding channels in reverse order to see if the channels show that way on DVR+
        # nope, DVR+ seems to sort them low to hi Ch #, based on OTA scan, as expected i guess.

        for x in LU['ChannelList']:
            
            if not x['favorite'] :  # if this is not a user favorite, we wont add it to our 'service' 
                continue            # continue on to next channel
            
            chstr = x['channel']    # get the channel string for use later
            
            # Next search for a matching callsign string like 'WISCDT' in stationlist so we can get the stationId  
            # Need to seach by callsign because there can be duplicate channel numbers in a Lineup
            # e.g 14.1 KBVO-DT
            #     14.1 KBVO-CD

            CallSign = x['callsign']

            # To Allow a Schedule to be placed on another Channel...
            
            # Alternate Schedule CallSign
            AltSchedule  = x.get('schedule', "")
            if AltSchedule :
                CallSign = AltSchedule
                logging.warning('BSL: Remapping Schedule for Channel %s to %s ...', x['channel'], AltSchedule )

            # Search for a matching SD schedule callsign
            Found = False
            for chn in stationList :
                if CallSign == chn['callsign'] :
                    Found = True
                    break;

            # Allow user to put a custom icon in /myicons for stations that dont have schedules
            if CallSign == 'MYICON' :
                logging.info('BSL: Ch %s  Encountered "MYICON" place image file %s.png in /myicons...', chstr, chstr )
                Found = True
            
            # Python loop var chn sticks at last list item so we cant test it for going past end of list
            
            if not Found :
                logging.warning('BSL: Callsign %s for Channel %s not found in StationList!', CallSign, x['channel'] )
                continue  # printed warning, we will skip b/c we cant add it if it does not exist

            # Get the stationID - for channels that dont exist, use stationID of '0'
            if CallSign == 'MYICON' :
                stationID = '0'
            else :
                stationID = chn.get('stationID')  # This is a string, (0 for MYICON)

            # Now that we found the correct station callsign... well, we are done,
            # we have the stationID and we have the Chan Number from the preview list,
            # but we can go to the Map section and get the uhfVhf for fun
            
            Found = False
            for mapChan in mapList :
                if mapChan['stationID'] == stationID :
                    Found = True
                    break;
                
            if not Found :
                logging.warning('BSL: StationID %s not found in MapList! Continuing', stationID )
                #continue  # printed warning, but we will continue b/c we have the data we need

            # get the one unique attibute from the maplist
            uhfVhf = mapChan.get('uhfVhf',0)   # 

            # Test for Analog channels, skip if found
            # Build a channel number string from the 2 supplied atsc integers
            # Analog channels dont have atsc keys, using get() is safe.
            # If atsc keys are missing, this wall make a chstr of ".", 
            achstr = str( mapChan.get("atscMajor","")  ) + "." + str( mapChan.get("atscMinor","") )

            if achstr == "." :
                logging.warning('BSL: Analog Channel selected.  Skipping channel %s (%s)!', x['channel'], x['callsign'] )
                continue

            # Test for duplicate channels, which can happen with Multiple Lineups.
            # UI was supposed to prevent this, but user could have edited thier pigs.json
            
            if x['channel'] not in ChannelList :    # keep a list of channel numbers encountered
                ChannelList.append( x['channel'] )
            else :
                logging.warning('BSL: ******* Skipping Duplicate Channel %s in Lineup: %s ******', x['channel'], LU['LineupName'] )
                continue  # printed warning, will skip b/c we dont want dup chan numbers, that would confuse DVR+

            # ---------------------------------------------------------------------
            # Made it this far - add the Channel to our Services list.
            
            # Add a new Channel object, recursive dict copy
            
            gServiceList.append( copy.deepcopy(ChannelDef) )  # add a Channel / Stations ID
            
            # My own custom key
            gServiceList[-1]["SDLineupName"] = LU["LineupName"]

            # To Remap schedules, DVR needs a unique SourceID for each station.
            # I will use my own Key 'ScheduleId' for the SD schedule and create a
            # 'unique' SourceId for DVR based on the channels position(order) in the service list
            
            # primary key for a channel is the stationID
            gServiceList[-1]["ScheduleId"]     = int(stationID)  # must match Airings Source ID

            # Create a unique SourceId for user re-mapped schedules to please DVR 
            if AltSchedule :
                gServiceList[-1]["SourceId"]   = int(stationID) + (1000*order)  # must match Airings Source ID
            else :
                gServiceList[-1]["SourceId"]   = int(stationID)  # must match Airings Source ID
                
            gServiceList[-1]['ChannelLocation'] = uhfVhf  # informational - stuff uhfVhf in unused Key
            gServiceList[-1]["ParentNetworkId"] = order  # give each station a different parentId just b/c
            
            # Merge what we know from the two SD lists
            # these next two have to match or DVR+ wont display schedule for this channel
            gServiceList[-1]["Channel"]             = copy.copy(chstr)
            gServiceList[-1]["VirtualChannelNumber"]= copy.copy(chstr)
            gServiceList[-1]["Order"]               = order  # Not sure if order does anything???
            gServiceList[-1]["DisplayName"]         = copy.copy(x['callsign']) # like WISCDT
            gServiceList[-1]["FullName"]            = copy.copy(x.get("name"))     # like WISC HTTV

            # appears that some SD stations dont have an affiliate key
            # we will just use the 'callsign' if 'affiliate' is not available
            affiliateCallLetters                    = x.get("affiliate")    # like CBS
            if affiliateCallLetters :
                gServiceList[-1]["CallLetters"]     = copy.copy(affiliateCallLetters)
            else :
                gServiceList[-1]["CallLetters"]     = copy.copy( x['callsign'] )
            
            gServiceList[-1]["IconAvailable"] = True   #images below
            
            # ============ Channel Icon Images ====================
            # Create an image structure.
            # Channel Icon Strategy.
            # 1) Look for a custom icon in ./myicons
            #    look in ./myicons/ for a icon file named <ChNum>.png, like 3.1.png
            #    If found, use it.  if not, then goto #2
            #    User can edit these icon files and put whatever they want.
            #    This allows a user to create their own channel Icons
            #    https://tvschedule.zap2it.com/?aid=gapzap has nice 55wx41h images
            #    DVR+ display area in guide is 92x51w pixels
            # 2) If 'IconType' ==2 ICON_ROVI on the UI:
            #    look for a cached(saved) CM/Rovi icon in ./cmicons/
            #       for a file named <ChNum>.png, like 3.1.png
            #    If found, use it, otherwise get the CM image and save it in ./cmicons as <ChNum>.png
            #    Now same as #1. only way to update cm icons would be to delete the png's in cmicons
            # 3) And lastly, if no user icon and if Rovi is turned off or not found, attempt to get
            #    an icon from Schedules Direct.
            #    SD Icons are resized and Cached in ./sdicons.  There is an option to save the FullSize Sd icon
            # 4) If nothing found, then use ['DefaultIcon'] pink pig icon
            #
            # Our goal is the get something into IconUrl to stick into the DVR services structure
            #
            # For myicons or cached cm icons stored on PiGS:
            #   icons are stored on Pi but DVR+ needs a full URL to make the request.
            #   The our DNS 'redirect' will send the icon request right back to PiGS

            IconUrl   = None    #default to dvr+ little TV icon if nothing found
            
            #----------------------------------------------------------------------------------
            # First check for a custom icon file in './myicons' for an icon file named x.y.png
            
            iconfname = 'myicons/' + chstr + '.png'

            # 1.40 Add user setting on UI for 'UseMyIcons' allow disable of myicons
            if Settings["UseMyIcons"] and os.path.exists(iconfname) :
                IconUrl = 'http://epg.channelmastertv.com/' + iconfname

            # if user selects MYICON, there 'must' be an icon in /myicons, pig if not found
            if not IconUrl and CallSign == 'MYICON' :
                iconfname = Settings.get('DefaultIcon', "pinkpig.png")
                IconUrl = 'http://epg.channelmastertv.com/' + iconfname

                
            #----------------------------------------------------------------------------------
            # if no icon set yet and user says Use Rovi Icons...
            
            if not IconUrl and Settings['IconType'] == ICON_ROVI:
                
                # See if we have a cached CM image icon in the icon cache directory
                iconfname = 'cmicons/' + chstr + '.png'  # name like cmicons/3.1.png

                if os.path.exists(iconfname) and not Settings["UpdateIconCache"] :
                    
                    IconUrl = 'http://epg.channelmastertv.com/' + iconfname
                    
                else :
                    # The Channel Master service lineup was retrieved above,
                    # search CM's list for the current channel.
                    # If we do not have an x.y.png icon file, try and get one from CM
                    
                    # Get CM's image structure, returns None if cant find matching channel
                    cmImageDef = CMGetChannelImageDef(chstr)
                        
                    if cmImageDef :  # check if we got an image structure from Channel Master
                        
                        logging.info('Getting CM channel icon url for channel %s', chstr )
                        
                        cmLogoUrl = cmImageDef['ImageUrl']
                        
                        # If Caching CM Icons, save a copy of CM's image on the Pi in ./cmicons
                        # https://kite.com/python/answers/how-to-download-an-image-using-requests-in-python
                        if Settings["CacheCMIcons"] :
                            response = requests.get(cmLogoUrl)
                            file = open(iconfname, "wb")
                            file.write(response.content)
                            file.close()
                            # Url to cached CM Icon on PiGS (since epg.xxx is redirected to PiGS)
                            IconUrl = 'http://epg.channelmastertv.com/' + iconfname
                        else :
                            # Go Out to CM in realtime for the Icons
                            IconUrl = cmLogoUrl

            #----------------------------------------------------------------------------------
            # If we dont have an IconUrl yet, try for a Schedules direct Icon
            #logging.info('cfg icontype %s %s', Settings['IconType'],  type(Settings['IconType']) )
            #logging.info('    icontype %s %s', ICON_SCHD, type(ICON_SCHD) )
            
            if not IconUrl and Settings['IconType'] == ICON_SCHD :
               
                # See if we have a cached SD image icon in the sdicon cache directory
                iconfname = 'sdicons/' + chstr + '.png'  # name like sdicons/3.1.png

                # if the file exists in the directory sdicons and Cache update not needed
                if os.path.exists(iconfname) and not Settings["UpdateIconCache"] :
                
                    IconUrl = 'http://epg.channelmastertv.com/' + iconfname
                    
                else :
                    
                    # "SdIconTheme"  : "dark", # choices are dark, gray, light, white
                    # get the list of SD Logos
                    logoList = chn.get("stationLogo", None)
                    
                    # Check if the channel actually has a list of images , 
                    if logoList :
                        # search for the theme the user requested
                        for logo in logoList :
                            if logo["category"] == Settings['SdIconTheme'] :
                                logging.debug('Matched SD icon theme: %s!', logo["category"] )
                                break
                        #wherever we are right now is the default user will get
                        #either we matched the icon theme/category or they
                        #get the last item in SD's list as a default
                        
                        SdIconUrl = logo["URL"] 
                        #SdIconUrl = logo["URL"] + "?token=" + token
                        
                        # ************  SD STARTS HERE *************************************
                        # If Caching SD Icons, save a copy of SD's image on the Pi in ./sdicons

                        if Settings["CacheSDIcons"] and PillowLib :  # need PillowLib to do resize
                    
                            logging.info('Fetching SD icon for %s...  %s', chstr, SdIconUrl )
                            
                            # Get FS SD icon, and Save raw full size as 3.1fs.png
                            response = requests.get(SdIconUrl)
                            
                            # img holds the full size image in memory
                            img = Image.open(BytesIO(response.content))
                            
                            # save a full size copy (overwrites i think)
                            if Settings["SaveFsSDIcons"] :
                                img.save('sdicons/' + chstr + 'fs.png', "PNG")
                            
                            # Pillow installation - 
                            # https://pillow.readthedocs.io/en/stable/installation.html#
                            # https://pypi.org/project/Pillow/#history
                            # here is the link to the library
                            # https://pillow.readthedocs.io/en/stable/reference/Image.html?highlight=resize#PIL.Image.Image.resize
                            # see Filters comparison table in next link LANCZOS
                            # https://pillow.readthedocs.io/en/stable/handbook/concepts.html#concept-filters
                            # note that pillow 5.2-5.4 supports python 3.5-3.7,
                            # pip3 list
                            # pip3 show pillow - mine is 5.4.1
                            
                            # Image.resize(size, resample=3, box=None, reducing_gap=None)
                            # resample codes
                            # Image.NEAREST  = 0
                            # Image.LANCZOS  = 1
                            # Image.BILINEAR = 2
                            # Image.BICUBIC  = 3   << default if nothing sent
                            # Image.BOX      = 4
                            # Image.HAMMING  = 5

                            # Note reducing_gap  and reduce() are Pillow version 7.0 features
                            # https://pillow.readthedocs.io/en/stable/releasenotes/7.0.0.html#
                            # we may need to make our min Python verions be 3.5 to support newer Pillow libs

                            logging.info('Resizing SD icon for channel %s...', chstr )
                            
                            # CM/Rovi ICON size is 92x36h
                            # DVR+ available screen size appears to be 92x51h
                            # SD icon size is 360x270,
                            # 360w x 270 / 10  = 36 x 27
                            # 360w x 270 / 9   = 40 x 30
                            # 360w x 270 / 6   = 60 x 45
                            # 360w x 270 / 5   = 72 x 54  <<<<<<
                            
                            # Get imaging params from config file
                            # convert json array to Tuple
                            Size_tuple   = ( Settings["SDImSize"][0], Settings["SDImSize"][1] )
                            ResampleType = Settings["SDImResample"]
                            ReducingGap  = Settings["SDImReduceGap"]
                            
                            if PillowVer < 7.0 :
                                img_resized = img.resize(Size_tuple, resample = ResampleType)
                            else :
                                img_resized = img.resize(Size_tuple, resample = ResampleType, \
                                                         box=None, reducing_gap=ReducingGap)
                            
                            #img_resized = img.resize((60, 45),resample = 1)  #exactly 6x reduction 
                            #img_resized.save('sdicons/' + chstr + '_1-60x45.png', "PNG")  # name 'sdicons/3.1_48x36.png'                            
                            
                            img_resized.save(iconfname, "PNG")  #this is the 3.1.png form
                            
                            # this line tells the DVR to find the resized image in the sdicons dir
 
                            IconUrl = 'http://epg.channelmastertv.com/' + iconfname
                            
                        else :
                            # Go Out to SD in realtime for Full Size Icons
                            # SD logo URL's have https:// in front of them but dvr+ doesnt seem to like this
                            
                            # so we will strip off the 'scheme'
                            o=urlparse(SdIconUrl)   # break url into scheme,netloc,path,params,query,fraqment
                            
                            # In Aug 2021 SD started to require 
                            # and just use http://  it seems to work ok at SD
                            IconUrl = 'http://' + o.netloc + o.path
                    else :                            
                        logging.warning('No SD icons for channel %s!', chstr )
                    
            #----------------------------------------------------------------------------------
            # At this point we should have an image url in IconUrl,
            # if we do not, we'll use the 'DefaultIcon' a pink pig

            if not IconUrl :
                iconfname = Settings.get('DefaultIcon', "pinkpig.png")

                IconUrl = 'http://epg.channelmastertv.com/' + iconfname
                                       
                # could also put up a random piggy from the list if we have nothing else
                # for testing, select 1 of 3 logos based on the order index
                #logos = ("http://epg.channelmastertv.com/pigoutline92x36.png",
                #         "http://epg.channelmastertv.com/misspiggy40x30.png",
                #         "http://epg.channelmastertv.com/pinkpig.png")
                # index = order % 3
                # IconUrl = logos[index]

            #----------------------------------------------------------------------------------
            # over-ride normal icon in one slot and put a piggy in the guide
            # order can never be 0, so setting pings.json 'PinkPigPos' to 0 disables over-ride
            # PinkPigPos can be an int row number in the channel line-up,
            # or a channel number like 3.1 or "3.1"
            pp_pos = False
            if isinstance(Settings['PinkPigPos'], int) :
                pp_pos = (order==Settings['PinkPigPos'])
            if isinstance(Settings['PinkPigPos'], float) :
                pp_pos = (chstr == str(Settings['PinkPigPos']))
            if isinstance(Settings['PinkPigPos'], str) :
                pp_pos = (chstr == Settings['PinkPigPos'])
                
            if pp_pos :
                logging.info("Over-riding channel Icon due to pigs.json['PinkPigPos']" )
                IconUrl = "http://epg.channelmastertv.com/pinkpig.png"

            # and now tell the log what icon was actally selected for this channel
            logging.info('BSL: Ch %s  Logo url %s', chstr, IconUrl)

            #----------------------------------------------------------------------------------
            # NOW CREATE THE IMAGE STRUCTURE
            # PiGS will provide an icon of some sort for every channel
            # get an 'empty' image object
            
            # As long is images are not disabled...
            if Settings['IconType'] > ICON_NONE :
                gServiceList[-1]["ChannelImages"].append( copy.copy(ImageDef) )   #copy.copy
                img0 = gServiceList[-1]["ChannelImages"][0]  # fix up one reference so icon displays
                img0['ObjectId'] = int(stationID)            # this must match for image to display
                img0["ImageUrl"] = copy.copy(IconUrl)        # Put icon url into service structure

            order += 1   # on to the next channel
            
            # end of channel loop  "for x in LU['ChannelList']:"

        # dispose of CM image list at end of each service
        # otherwise the CM module will hold a big variable in memory
        
        CMDisposeServcies()
        logging.info('BSL: Lineup %s Processed...', LU['LineupName'] )

        # end of Lineup Loop   "for LU in Settings['MyLineups']"


    logging.info('BSL: Build complete. %d stations in Service list.', order-1)

    # update last build times
    lt,lz = LocalTimeStr(datetime.utcnow(), True)

    # update module status
    SvcBuildInfo[ServicesGrpHeader][BUILDKEY] = lt
    SvcBuildInfo[ServicesGrpHeader][CHAN_KEY] = order-1

    # v1.40, save service list to disk
    if Settings['SaveSchedule'] == 2 :
        logging.info('BSL: Saving Services List to Disk...')
        utSaveObjectToDisk(gServiceList, SERVICE_FILE)
   
    return gServiceList

#=====================================================
# This is the handling of DVR+ Message Type 2
# create a response that will make DVR+ happy

def GetServiceDetails(Settings, ServiceID, tzlist):
                         
    global gServiceList
    
    theTime=datetime.utcnow()

    # Update the json Response Header with the current time and pack in the 99999 service id
    # Most of the stuff in the header is ignored by DVR+
    
    sdh = ServiceDetailHeader["ServiceDetailsResult"]  #to make the next lines shorter/cleaner
    
    sdh["ChannelLineup"]["ServiceId"] = ServiceID
    sdh["TimeStamp"]                  = theTime.strftime("%Y-%m-%dT%H:%M:%S.000Z")
    sdh["ChannelLineup"]["Name"]      = "The Pig Farm - Broadcast TV"
    
    sdh["ChannelLineup"]["TimeZones"] = copy.deepcopy(tzlist)

    #sdh["ChannelLineup"]["Name"]= Settings["metadata"]["lineup"]  #SD name of lineup
    #sdh["ChannelLineup"]["ServiceClass"]=Settings["metadata"]["transport"]
    
    # get reference to the channel list in the message block
    #clist = ServiceDetailHeader["ServiceDetailsResult"]["ChannelLineup"]["Channels"]
    clist = sdh["ChannelLineup"]["Channels"]
    
    clist.clear()  # clear out the prior Channel Lineup before adding the current lineup
    
    # add the global dict to the message block
    clist.extend(gServiceList)
    
    #convert Python dict to json and return js
    js = json.dumps(ServiceDetailHeader, indent=2)
    
    return js
